####################################################################################################
# # Copyright (C) 2024-Present - Daniel Charytonowicz - All Rights Reserved
# # Written by: Daniel Charytonowicz
# # Contact: daniel.charytonowicz@icahn.mssm.edu
# ###################################################################################################

from typing import Union, Tuple, Iterable

from anndata import AnnData, read_h5ad
import scipy.sparse.linalg as la
from scipy.sparse import issparse

import networkx as nx
import numpy as np
import pandas as pd
import scanpy as sc
import os


def load_ppi(
    ppi : str
) -> AnnData:
    """\
    Load PPI
    
    Loads a protein-protein interaction dataset based on string identifier.
    
    Params
    -------
    ppi
        Name of protein-protein interaction dataset to use. Loaded
        in format 'ppi_{ppi}_pp_sparse.h5ad'
    
    Returns
    -------
    ppi : anndata.AnnData
        The ppi matrix as a sparse annotated dataset
    """
    
    # Read filepath from local context and then load the corresponding PPI
    filepath = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            f"data/ppi_{ppi}_pp_sparse.h5ad")
    return read_h5ad(filepath)


def preprocess(
    data : AnnData,
    ppi : AnnData,
) -> Tuple[AnnData, AnnData]:
    """\
    Data Input and PPI Preprocessing
    
    Converts provided PPI into a fully-connected subgraph required for
    entropy calculation, and re-indexes both PPI and data to match
    shared gene symbols.
    
    Params
    -------
    data
        The annotated data matrix.
    ppi
        The PPI matrix as an annotated data object.
    
    Returns
    -------
    data : anndata.AnnData
        The annotated data matrix with reduced gene symbols to match
        gene symbol intersection with PPI
    ppi : anndata.AnnData
        The protein-protein interaction matrix with genes that
        represent the largest connected subgraph, and those that
        intersect with the input data matrix.
    """
    
    # Find common gene intersection between the two objects
    common_genes = pd.Index(np.intersect1d(data.var_names, ppi.var_names))
    
    # Ensure that there is sufficient overlap
    assert len(common_genes) > 1000, "Low number of genes overlapping reference PPI and target data"
    
    # Re-index keeping only common genes
    ppi = ppi[common_genes, common_genes]
    data = data[:, common_genes]
    
    # use networkX to find the largest connected subgraph
    gr = nx.Graph(zip(ppi.X.tocoo().row, ppi.X.tocoo().col))
    gr = nx.relabel_nodes(gr, dict(enumerate(ppi.var_names)))
    largest_clust_genes = sorted(max(nx.connected_components(gr), key=len))
    
    # Re-index keeping only largest cluster genes
    ppi = ppi[largest_clust_genes, largest_clust_genes]
    data = data[:, largest_clust_genes].copy()
    
    # Intvert data
    if data.X.max() < 30:
        print("Data is log-scaled, inverting")
        data.X = np.expm1(data.X)
    
    print("Normalizing data")
    sc.pp.normalize_total(data, 1e4)
    sc.pp.log1p(data)

    return data, ppi


def calc_max_entropy(
    ppi : AnnData
) -> float:
    """\
    Calculate maximum entropy.
    
    Calculates maximal entropy given a PPI matrix as a
    sparse anndata object. Calculated by taking the
    log of the right real eigenvalue.
    
    Params
    -------
    ppi
        The PPI matrix as an annotated data object.
    
    Returns
    -------
    max_entropy : float
        Maximum entropy of the given ppi matrix
    """
    
    # Calculate maximum entropy as log of right real eigenvalue
    eig_val, _ = la.eigs(ppi.X, k=1, which='LM')    
    return np.log(float(eig_val.real))


def chunk(
    obj : Iterable,
    chunksize : int = None
) -> Iterable:
    """\
    Iterable chunking
    
    Chunks an iterable (obj) into blocks of chunksize, 
    if the length of the last yielded value of this
    generaotr is smaller than the chunksize, then
    it will return a smaller chunk.
    
    Params
    -------
    obj
        Iterable-like object that can be chunked and
        has a length
    chunksize
        size of chunks to return
        
    Returns
    -------
    chunk : Iterable
        chunk of the underlying iterable
        
    """
    
    # Get chunksize, if not default to length of object
    chunksize = chunksize if chunksize else len(obj)
    
    # Return block of iterable object in chunks
    for i in range(0, len(obj) if not issparse(obj) else obj.shape[0], chunksize):
        yield obj[i : i + chunksize]



####################################################################################################
# # Copyright (C) 2024-Present - Daniel Charytonowicz - All Rights Reserved
# # Written by: Daniel Charytonowicz
# # Contact: daniel.charytonowicz@icahn.mssm.edu
# ###################################################################################################

import torch
from torch.utils.data import Dataset
from anndata import AnnData
from scipy.sparse import issparse
import numpy as np

class EntropyDataset(Dataset):
    """
    EntropyDataset
    fafaf
    A Dataloader for anndata objects
    
    Params
    -------
    adata
        Annotated dataset comprising a preprocessed dataset to compute entropy on.
    """
    
    def __init__(self, adata : AnnData):
        self.data = adata.X
                
    def __len__(self):
        return self.data.shape[0]
    
    def __getitem__(self, idx):
        
        # Get vector at index
        exp = self.data[idx]
        
        # convert from sparse to dense if desired
        exp = exp.toarray() if issparse(exp) else exp
        
        # Flatten
        exp = exp.flatten()
        
        # Cast to tensor and yield
        return torch.tensor(exp)
        



####################################################################################################
# # Copyright (C) 2024-Present - Daniel Charytonowicz - All Rights Reserved
# # Contact: daniel.charytonowicz@icahn.mssm.edu
# ###################################################################################################

from setuptools import setup, find_packages
from pkg_resources import parse_requirements
import pathlib

# Read requirements text file to get required packages
with pathlib.Path('requirements.txt').open() as reqsfile:
    reqs = [str(req) for req in parse_requirements(reqsfile)]

# Setup package with params
setup(name='pysce',
      version='0.1.2',
      description='pySCE: Single Cell Entropy Scoring in Python',
      url='https://github.com/dchary/pysce',
      author='Daniel Charytonowicz',
      author_email='daniel.charytonowicz@icahn.mssm.edu',
      license='GNU GPLv3',
      install_requires=reqs,
      packages=find_packages(),
      package_data={'pysce': ['data/ppi_scent_pp_sparse.h5ad']},
      include_package_data = True,
      zip_safe=False)


####################################################################################################
# # Copyright (C) 2023-Present - Daniel Charytonowicz - All Rights Reserved
# # Written by: Daniel Charytonowicz
# # Contact: daniel.charytonowicz@icahn.mssm.edu
# ###################################################################################################

from typing import Union, Tuple, Optional

from ._utils import load_ppi, preprocess, calc_max_entropy, chunk
from ._dataloaders import EntropyDataset
from ._modules import EntropyModule

from torch.utils.data import Dataset, DataLoader

from anndata import AnnData
import pandas as pd
import numpy as np
import torch

from tqdm.auto import tqdm
    


def score(
    data : Union[AnnData, pd.DataFrame],
    ppi : str = 'scent',
    use_raw : bool = True,
    batch_size : Optional[int] = 8,
    key_added : str = 'entropy',
    artifact_genes = ('RPS','RPL','MT'),
    inplace : bool = True,
    layer : Optional[str] = None,
) -> Optional[AnnData]:
    """
    Score Entropy
    
    Scores entropy for each cell in annotated dataset.
    
    Params
    -------
    data
        Annotated dataset comprising a preprocessed dataset to compute entropy on.
        Can also be a dataframe.
    ppi
        The ppi network to use, currently only 'scent' is supported.
    use_raw
        Whether to use 'adata.raw' for counts source.
    batch_size
        Size of batch to compute entropy with, default is 8 cells. If you experiences
        out of memory errors, try decreasing this value. Increase this value to increase
        GPU utilization, but note memory issues may arise.
    key_added
        The key in 'adata.obs' to add with entropy values computed.
    artifact_genes
        Genes to exclude from entropy calculation.
    inplace
        Compute changes to data inplace
    layer
        Whether to get counts source from a layer in data.
        
    Returns
    -------
    data or None if in-place
    
    """
    
    # Check inputs validity
    assert isinstance(data, (AnnData, pd.DataFrame)), "Input data must be AnnData or DataFrame object."
    assert ppi == 'scent', "PPI must be scent"
    assert batch_size >= 1 if batch_size else True, "Batch size must be >= 1 or None (auto-calculate)"
    
    # Create a reference to the original data object
    data_orig = data
    
    # If passing a dataframe, convert to anndata object
    is_adata = isinstance(data, AnnData)
    data = data if is_adata else AnnData(data)
    
    # If raw exists and is to be used, pass
    if not layer:
        data = data.raw.to_adata() if (use_raw and data.raw) else data
    else:
        data = AnnData(data.layers[layer], obs = data.obs, var = data.var)
    
    # Remove artifact genes
    data = data[:, data.var[~data.var.index.str.startswith(artifact_genes)].index] if artifact_genes else data

    # Load PPI using provided name
    ppi = load_ppi(ppi)
    
    # Preprocess data, PPI, and calculate maximum entropy
    data, ppi = preprocess(data, ppi)
    
    # Calculate maximum entropy as log of right real eigenvalue
    max_entropy = calc_max_entropy(ppi)
    
    # setting device on GPU if available, else CPU
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    # Get dataset
    dataset = EntropyDataset(data)

    # Create dataloader
    dataloader = DataLoader(dataset, batch_size = batch_size)
    
    # Create entropy module
    module = EntropyModule()

    # Assign to devices if available
    module = torch.nn.DataParallel(module, device_ids=[0])
    
    # Make PPI tensor
    ppi_tensor = torch.tensor(ppi.X.toarray()).to(device)
    
    # Buffer to hold scores
    entropy_scores = []

    # Create a progressbar to monitor progress
    with tqdm(total=len(dataloader), unit='batch') as pbar:

        pbar.set_description(f"Scoring Entropy")

        # Iterate each batch in dataloader
        for i, batch in enumerate(dataloader):

            # Push batch into target device
            batch = batch.to(device)

            # Compute entropy
            entropy = module(batch, ppi_tensor, max_entropy)

            # Add to results buffer
            entropy_scores.append(entropy.cpu().numpy())

            # Update progressbar
            pbar.update()
    
    # Merge scores
    entropy_scores = np.concatenate(entropy_scores)
    
    # Add key
    data_orig.obs[key_added] = entropy_scores
    
    # Return if not inplace
    if not inplace:
        return data_orig



####################################################################################################
# # Copyright (C) 2024-Present - Daniel Charytonowicz - All Rights Reserved
# # Written by: Daniel Charytonowicz
# # Contact: daniel.charytonowicz@icahn.mssm.edu
# ###################################################################################################

import torch

class EntropyModule(torch.nn.Module):
    """\
    
    Entropy Scoring Module
    
    This is the base module that contains an entropy scoring forward function.
    
    Params
    -------
    
    """
    def __init__(self):
        super().__init__()
        
    @torch.compile
    def forward(self, 
        exp : torch.tensor, 
        ppi : torch.tensor,
        max_entropy : float, 
        expand : bool = False
        ) -> torch.tensor:
        """\
        Forward Function To Calculate Entropy Score
        
        """
        
        # If batch size is 1 and expression is a vector then expand dimensions
        exp = exp.unsqueeze(0) if expand else exp

        # Transpose expression tensor to shape (genes, batch_size)
        exp = exp.t()

        # Calculate stationary distribution used later when computing
        # the final Markov chain entropy
        m_exp = exp * torch.matmul(ppi, exp)
        m_exp = m_exp / m_exp.sum(0)

        # Map cell expression onto the PPI
        pm = ppi * (exp.unsqueeze(0).permute([2, 1, 0]) * exp.unsqueeze(0).permute([2, 0, 1]))

        # Normalize each column to create a markov chain
        pm = torch.nan_to_num(pm / pm.sum(1, keepdim = True))

        # Calculate entropy score
        entropy = -torch.sum(torch.nan_to_num(pm.log() * pm), 1, keepdim = True)
        entropy = torch.reshape(entropy, shape = [entropy.shape[0], entropy.shape[2]]).t()
        entropy = torch.sum(m_exp * entropy, 0, keepdim = True)

        # Normalize entropy
        entropy = torch.squeeze(entropy / max_entropy)

        return entropy
